#!/usr/bin/perl

use strict;

use XML::DOM;
use XML::DOM::XPath;
use Getopt::Long;

XML::DOM::ignoreReadOnly (1);

my %OPT;
$OPT{diff} = "";
$OPT{prefix} = "named_entity_man";

sub usage {
    print "Usage: $0 [OPTIONS] FILE FILE\n";
    print "\n";
    print "  --diff=FILE     \tcreate a alvis-ne-diff\n";
    print "  --prefix=STRING \tset the prefix for new named entities\n";
}

if (grep(/help|aide| -h/,@ARGV)) {
    usage();
    exit;
}

GetOptions(
	   "diff=s" => \$OPT{diff},
	   "prefix=s" => $OPT{prefix}
	   );

$OPT{reference} = shift;
$OPT{manual} = shift;

my @nodes;
my $node;

# reference file
my $reference_parser = new XML::DOM::Parser;
my $reference_doc = $reference_parser->parsefile($OPT{reference});
my $reference_root = $reference_doc->getDocumentElement;

# manual annotation
my $manual_parser = new XML::DOM::Parser;
my $manual_doc = $manual_parser->parsefile($OPT{manual});
my $manual_root = $manual_doc->getDocumentElement;

# new document
my $new_document = new XML::DOM::Document;
my $new_root = $reference_root->cloneNode;
$new_root->setOwnerDocument($new_document);
$new_document->insertBefore($new_root);

# diff document
my $diff_document;
my $diff_root;

if ($OPT{diff} ne "") {
    $diff_document = new XML::DOM::Document;
    $diff_document->setDoctype($diff_document->createDocumentType("diff", "alvis-ne-diff.dtd"));
    $diff_root = new XML::DOM::Element($diff_document, "diff");
    $diff_document->insertBefore($diff_root);
}

@nodes = $reference_root->findnodes("documentRecord");
$node = shift @nodes;
my $new_dr = $node->cloneNode;
$new_dr->setOwnerDocument($new_document);
$new_root->insertBefore($new_dr);

@nodes = $reference_root->findnodes("documentRecord/acquisition");
$node = shift @nodes;
$node->setOwnerDocument($new_document);
$new_dr->insertBefore($node);

@nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis");
$node = shift @nodes;
my $new_la = $node->cloneNode;
$new_la->setOwnerDocument($new_document);
$new_dr->insertBefore($new_la);

foreach my $level (("lemma_level", "log_level", "morphosyntactic_features_level")) {
    @nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis/$level");
    $node = shift @nodes;
    if (defined $node) {
	$node->setOwnerDocument($new_document);
	$new_la->insertBefore($node);
    }
}

@nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis/semantic_unit_level");
$node = shift @nodes;
my $new_sul;
if (defined $node) {
    $new_sul = $node->cloneNode;
    $new_sul->setOwnerDocument($new_document);
}
else {
    $new_sul = $new_document->createElement("semantic_unit_level");
}
$new_la->insertBefore($new_sul);

foreach $node ($reference_root->findnodes("documentRecord/linguisticAnalysis/semantic_unit_level/semantic_unit[term]")) {
    $node->setOwnerDocument($new_document);
    $new_sul->insertBefore($node);
}

my $txtpos = 0;
my $manid = 1;
my @touched_ne = ();
@nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis/token_level");
my $reference_tl = shift @nodes;
@nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis/semantic_unit_level");
my $reference_sul = shift @nodes;

@nodes = $manual_root->findnodes("/document/text");
my $man_text = shift @nodes;
if (defined $man_text) {
    foreach my $man_node ($man_text->getChildNodes) {
	if ($man_node->getNodeType == TEXT_NODE) {
	    $txtpos += length($man_node->getNodeValue);
	}
	else {
	    my %overlap_old = ();
	    my $new_ne_form = $man_node->getFirstChild->getNodeValue;
	    my $new_ne_type = $man_node->getAttribute("type");

	    # Create the new semantic_unit node
	    my $new_su = new XML::DOM::Element($new_document, "semantic_unit");
	    $new_sul->insertBefore($new_su);
	    my $new_ne = new XML::DOM::Element($new_document, "named_entity");
	    $new_su->insertBefore($new_ne);
	    $node = new XML::DOM::Element($new_document, "form");
	    $new_ne->insertBefore($node);
	    $node->insertBefore(new XML::DOM::Text($new_document,$new_ne_form));
	    my $new_ne_id = new XML::DOM::Element($new_document, "id");
	    $new_ne->insertBefore($new_ne_id);
	    my $new_lrt = new XML::DOM::Element($new_document, "list_refid_token");
	    $new_ne->insertBefore($new_lrt);
	    $node = new XML::DOM::Element($new_document, "named_entity_type");
	    $node->insertBefore(new XML::DOM::Text($new_document, $new_ne_type));
	    $new_ne->insertBefore($node);

	    my @new_ne_tokens = $reference_tl->findnodes("token[from >= $txtpos and ".($txtpos+length($new_ne_form) - 1)." > from]/id");
	    foreach my $tokid (@new_ne_tokens) {
		my $tokid_text = $tokid->getFirstChild->getNodeValue;

		if (defined $reference_sul) {
		    @nodes = $reference_sul->findnodes("semantic_unit/named_entity[list_refid_token/refid_token = '$tokid_text']/id");
		}
		my $oldne = shift @nodes;
		if ($oldne) {
		    $overlap_old{$oldne->getFirstChild->getNodeValue} += 1;
		}
		$node = new XML::DOM::Element($new_document, "refid_token");
		$node->insertBefore(new XML::DOM::Text($new_document, $tokid_text));
		$new_lrt->insertBefore($node);
	    }
	    $txtpos += length($new_ne_form);

	    my $new_id;
	    my @local_touched_ne = keys %overlap_old;
	    my $first_ne = $local_touched_ne[0];
	    my @first_ne_tokens = ();
	    my $first_ne_type;
	    if (defined $reference_sul) {
		@first_ne_tokens = $reference_sul->findnodes("semantic_unit/named_entity[id = '$first_ne']/list_refid_token/refid_token");
		@_ = $reference_sul->findnodes("semantic_unit/named_entity[id = '$first_ne']/named_entity_type");
		$first_ne_type = shift @_;
	    }
	    if (($#local_touched_ne == 0) && ($#first_ne_tokens == ($overlap_old{$first_ne} - 1)) && ($#first_ne_tokens == $#new_ne_tokens) && ($new_ne_type eq $first_ne_type->getFirstChild->getNodeValue)) {
		$new_id = $first_ne;
	    }
	    else {
		$new_id = $OPT{prefix} . $manid;
		$manid += 1;
		if ($OPT{diff} ne "") {
		    my $diff_added = new XML::DOM::Element($diff_document, "added");
		    $diff_added->setAttribute("id",$new_id);
		    foreach my $n (@local_touched_ne) {
			my $diff_overlap = new XML::DOM::Element($diff_document, "overlap");
			$diff_overlap->setAttribute("id", $n);
			$diff_added->insertBefore($diff_overlap);
		    }
		    $diff_root->insertBefore($diff_added);
		}
	    }
	    $new_ne_id->insertBefore(new XML::DOM::Text($new_document, $new_id));
	    push @touched_ne, @local_touched_ne;

	    $new_sul->insertBefore($new_su);
	}
    }
}

foreach my $level (("sentence_level", "token_level", "word_level")) {
    @nodes = $reference_root->findnodes("documentRecord/linguisticAnalysis/$level");
    $node = shift @nodes;
    if (defined $node) {
	$node->setOwnerDocument($new_document);
	$new_la->insertBefore($node);
    }
}

$new_document->printToFileHandle(\*STDOUT);

if ($OPT{diff} ne "") {
    if (defined $reference_sul) {
	foreach $node ($reference_sul->findnodes("semantic_unit/named_entity/id")) {
	    my $neid = $node->getFirstChild->getNodeValue;
	    if (!grep(/^$neid$/,@touched_ne)) {
		my $diff_removed = new XML::DOM::Element($diff_document, "removed");
		@_ = $node->findnodes("../list_refid_token/refid_token[1]");
		my $token1 = shift @_;
		$token1 = $token1->getFirstChild->getNodeValue;
		@_ = $new_root->findnodes("documentRecord/linguisticAnalysis/word_level/word[list_refid_token/refid_token = '$token1']/id");
		my $wid = shift @_;
		$diff_removed->setAttribute("id", $neid);
		if (defined $wid) {
		    $diff_removed->setAttribute("word_id", $wid->getFirstChild->getNodeValue);
		}
		$diff_root->insertBefore($diff_removed);
	    }
	}
    }
    $diff_document->printToFile($OPT{diff});
}
