#!/usr/bin/perl

use strict;
use XML::DOM;
use XML::DOM::XPath;
use Getopt::Long;

my %OPT;
$OPT{header} = 0;
$OPT{data} = 0;
$OPT{count} = 0;

$OPT{vocabulary} = 0;
$OPT{following} = "";
$OPT{preceding} = "";
$OPT{full} = 0;
$OPT{diff} = 0;
$OPT{dict} = "";
$OPT{anti} = "";
$OPT{sink} = "";
$OPT{negative} = "";
$OPT{title} = 0;
$OPT{range} = 5;
$OPT{netype} = "";

sub usage {
    print "Usage: $0 [ACTION] [OPTIONS] FILE\n";
    print "Builds ARFF files for biology NE recognition.\n";
    print "\nActions:\n";
    print "  --header       \tprint ARFF header\n";
    print "  --data         \tprint data\n";
    print "\nOptions:\n";
    print "  --vocabulary   \tvocabulary mode\n";
    print "  --following    \tfollowing words attributes\n";
    print "  --preceding    \tpreceding words attributes\n";
    print "  --full         \tfull test set\n";
    print "  --diff         \tdiff file\n";
    print "  --dict=FILE    \tdictionary file\n";
    print "  --anti=FILE    \tfile containing negative examples\n";
    print "  --sink=FILE    \tfile containing sink words\n";
    print "  --negative=REXP\tregular expression matching negative examples\n";
    print "  --range=INT    \tnumber of words in the context\n";
    print "  --type=TYPE    \ttype of the named entities to process\n";
    print "  --title        \ttext is highlighted\n";
}

if (grep(/help|aide| -h/,@ARGV)) {
    usage();
    exit;
}

GetOptions(
	   "header" => \$OPT{header},
	   "data" => \$OPT{data},
	   "count" => \$OPT{count},

	   "vocabulary" => \$OPT{vocabulary},
	   "following=s" => \$OPT{following},
	   "preceding=s" => \$OPT{preceding},
	   "full" => \$OPT{full},
	   "diff" => \$OPT{diff},
	   "dict=s" => \$OPT{dict},
	   "anti=s" => \$OPT{anti},
	   "sink=s" => \$OPT{sink},
	   "negative=s" => \$OPT{negative},
	   "range=i" => \$OPT{range},
	   "type=s" => \$OPT{netype},
	   "title" => \$OPT{title}
	   );

my $CLASS_NAME = "class";
my $CLASS_REF = "{ne,not_ne}";

my %ATT_REF;
my @ATT_NAME = ("title", "first_upper", "middle_upper", "only_upper", "last_digit", "first_dash", "middle_dash", "paren", "space", "length", "between_paren", "eq_dict", "in_dict", "eq_anti", "in_anti");

sub file_to_array {
    my $fn = shift;
    my @a = ();

    open F, "<", $fn;
    while (my $l = <F>) {
	chomp $l;
	push @a, $l;
    }
    close F;

    return @a;
}

sub escape_re_chars {
    my $s = shift;
    $s =~ s/([\*\+\[\]\(\)\.\?\{\}\'])/\\$1/g;
    return $s;
}

sub positives {
    my $type = shift;
    my $root = shift;

    if ((defined $type) && ($type ne "")) {
	return $root->findnodes("//ne[\@type='$type']");
    }
    else {
	return $root->findnodes("//ne");
    }
}

sub negatives {
    my $full = shift;
    my $neg_form = shift;
    my @removed = @{shift()};
    my @anti = @{shift()};
    my @sink = @{shift()};
    my $root = shift;

    if ($full) {
	return $root->findnodes("//word[\@category = 'N']");
    }

    my @neg = ();
    foreach my $word ($root->findnodes("//word")) {
	my $is_neg = 0;
	my $wf = $word->getFirstChild->toString;

	my $aid = $word->getAttribute("alvis-id");
	if (grep(/^$aid$/,@removed))  {
	    $is_neg = 1;
	}
 	elsif ((defined $neg_form) && ($neg_form ne "") && ($wf =~ /$neg_form/)) {
	    $is_neg = 1;
 	}
	else {
	    $wf = escape_re_chars($wf);
	    my $pos = $word->getAttribute("category");
	    if (($pos =~ /^[NJ]/) && (grep(/^$wf$/,@anti))) {
		$is_neg = 1;
	    }
	    else {
		my @nn = $word->findnodes("following-sibling::word[1]");
		my $next = shift @nn;
		if ($next) {
		    my $nf = escape_re_chars($next->getAttribute("lemma"));
		    if (grep(/^$nf$/,@sink)) {
			$is_neg = 1;
		    }
		}
	    }
	}
	if ($is_neg) {
	    push @neg, $word;
	}
    }

    return @neg;
}

sub neighbours {
    my $range = shift;
    my $direction = shift;
    my $node = shift;
    my @neigh = ();

    if ($direction eq "after") {
	$direction = "following";
	print STDERR "WARNING: using 'after'\n";
    }
    if ($direction eq "before") {
	$direction = "preceding";
	print STDERR "WARNING: using 'before'\n";
    }
    @neigh = $node->findnodes("$direction-sibling::*[name() = 'word' or name() = 'ne']");;
    splice @neigh, $range;

    return @neigh;
}

sub vocabulary_data {
    my $class = shift;
    my $range = shift;
    my $example = shift;
    my @neigh = ();

    print "% ".$example->getFirstChild->toString."\n";
    foreach my $dir (("following", "preceding")) {
	@neigh = neighbours($range,$dir,$example);
	foreach my $i (1 .. $range) {
	    my $n = shift @neigh;
	    if (defined $n) {
		my $str;
		if ($n->getNodeName eq "word") {
		    if ($n->getAttribute("category") =~ /^[NJV]/) {
			$str = $n->getAttribute("lemma");
		    }
		}
		else {
		    $str = $n->getFirstChild->toString;
		}
		if ($str ne "") {
		    $str =~ s/\'/\\\'/g;
		    print "$i,$dir,'$str',$class\n";
		}
	    }
	}
    }
}

sub example_data {
    my $class = shift;
    my $range = shift;
    my $title = shift;
    my @dict = @{shift()};
    my @anti = @{shift()};
    my @following = @{shift()};
    my @preceding = @{shift()};
    my $example = shift();
    my $example2 = shift();
    my $example3 = shift();

    my $last_example;
    my $form;
    if (defined $example3) {
	$last_example = $example3;
	$form = join(" ", ($example->getFirstChild->toString, $example2->getFirstChild->toString, $example3->getFirstChild->toString));
    }
    elsif (defined $example2) {
	$last_example = $example2;
	$form = join(" ", ($example->getFirstChild->toString, $example2->getFirstChild->toString));
    }
    else {
	$last_example = $example;
	$form = $example->getFirstChild->toString;
    }


    my %attr = ();
    my @nodes = ();
    my $node = undef;

    $attr{title} = int($title);


    # NE internal form attributes
    $attr{first_upper}  = int($form =~ /^[A-Z]/);
    $attr{middle_upper} = int($form =~ /^.+[A-Z]/);
    $attr{only_upper}   = int($form =~ /^[A-Z]*$/);
    $attr{last_digit}   = int($form =~ /[0-9]$/);
    $attr{first_dash}   = int($form =~ /^-/);
    $attr{middle_dash}  = int($form =~ /^.+-/);
    $attr{paren}        = int($form =~ /\(.*\)/);
    $attr{space}        = int($form =~ / /);
    $attr{length}       = length($form);

    # Between parentheses attribute
    $attr{between_paren} = 0;
    $node = $example->getPreviousSibling;
    if ((defined $node) && ($node->getNodeType == TEXT_NODE) && ($node->getNodeValue =~ /\([ ]*$/)) {
	$node = $last_example->getNextSibling;
	if ((defined $node) && ($node->getNodeType == TEXT_NODE) && ($node->getNodeValue =~ /^[ ]*\)/)) {
	    $attr{between_paren} = 1;
	}
    }

    # Dictionnary attributes
    my $fre = escape_re_chars($form);
    $attr{eq_dict} = int(grep(/^$fre$/, @dict));
    if (grep(/.+$fre/, @dict)) {
	$attr{in_dict} = 1;
    }
    elsif (grep(/$fre.+/, @dict)) {
	$attr{in_dict} = 1;
    }
    else {
	$attr{in_dict} = 0;
    }

    $attr{eq_anti} = int(grep(/^$fre$/, @anti));
    if (grep(/.+$fre/, @anti)) {
	$attr{in_anti} = 1;
    }
    elsif (grep(/$fre.+/, @anti)) {
	$attr{in_anti} = 1;
    }
    else {
	$attr{in_anti} = 0;
    }

    # POS context attributes
    foreach my $dir (("following", "preceding")) {
	my @context;

	if ($dir eq "following") {
	    @context = @following;
	    @nodes = neighbours($range,$dir,$last_example);
	}
	else {
	    @context = @preceding;
	    @nodes = neighbours($range,$dir,$example);
	}
	foreach my $w (@context) {
	    $attr{$dir."_word_$w"} = 0;
	}

	for (my $i = 1; $i <= $range; $i++) {
	    $node = shift @nodes;
	    if (defined $node) {
		my $wf = $node->getAttribute("lemma");
		my $wef = escape_re_chars($wf);
		if (grep(/^$wef$/i,@context)) {
		    $attr{$dir."_word_$wf"} = 1;
		}
		if ($node->getNodeName eq "word") {
		    $attr{"pos_".$dir."_$i"} = $node->getAttribute("category");
		}
		elsif ($node->getNodeName eq "ne") {
		    $attr{"pos_".$dir."_$i"} = "N";
		}
		else {
		    print STDERR "Got ".$node->getNodeName.", expected word or ne\n";
		    exit 1;
		}
	    } else {
		$attr{"pos_".$dir."_$i"} = "?";
	    }
	}
    }

    print "% $form\n";
    foreach my $a (@ATT_NAME) {
	print $attr{$a}.",";
    }
    print "$class\n";
}

$ATT_REF{title} = "{0,1}";
$ATT_REF{first_upper} = "{0,1}";
$ATT_REF{middle_upper} = "{0,1}";
$ATT_REF{only_upper} = "{0,1}";
$ATT_REF{last_digit} = "{0,1}";
$ATT_REF{first_dash} = "{0,1}";
$ATT_REF{middle_dash} = "{0,1}";
$ATT_REF{paren} = "{0,1}";
$ATT_REF{space} = "{0,1}";
$ATT_REF{length} = "NUMERIC";
$ATT_REF{between_paren} = "{0,1}";
$ATT_REF{eq_dict} = "{0,1}";
$ATT_REF{in_dict} = "{0,1}";
$ATT_REF{eq_anti} = "{0,1}";
$ATT_REF{in_anti} = "{0,1}";

foreach my $dir (("following", "preceding")) {
    for (my $i = 1; $i <= $OPT{range}; $i += 1) {
	push @ATT_NAME, "pos_".$dir."_$i";
	$ATT_REF{"pos_".$dir."_$i"} = "{J,N,PP,RB,V,O}";
    }
}

my @following = ();
my @preceding = ();
if ($OPT{following}) {
    open F, "<", $OPT{following};
    while (my $l = <F>) {
	chomp $l;
	push @following, $l;
	push @ATT_NAME, "following_word_$l";
	$ATT_REF{"following_word_$l"} = "{0,1}";
    }
    close F;
}
if ($OPT{preceding}) {
    open P, "<", $OPT{preceding};
    while (my $l = <P>) {
	chomp $l;
	push @preceding, $l;
	push @ATT_NAME, "preceding_word_$l";
	$ATT_REF{"preceding_word_$l"} = "{0,1}";
    }
    close F;
}

if ($OPT{header}) {
    if ($OPT{vocabulary}) {
	print "\@RELATION vocabulary

\@ATTRIBUTE position NUMERIC
\@ATTRIBUTE direction {preceding,following}
\@ATTRIBUTE word STRING
\@ATTRIBUTE class {ne,not_ne}

\@DATA
";
    }
    else {
	print "\@RELATION ne_or_not\n\n";
	foreach my $a (@ATT_NAME) {
	    print "\@ATTRIBUTE $a ".$ATT_REF{$a}."\n";
	}
	print "\@ATTRIBUTE $CLASS_NAME $CLASS_REF\n";
	print "\n\@DATA\n";
    }
}

if ($OPT{data} || $OPT{count}) {
    my %count_examples;
    $count_examples{positive} = 0;
    $count_examples{negative} = 0;

    my @anti=();
    if ($OPT{anti}) {
	@anti = file_to_array($OPT{anti});
    }

    my @dict=();
    if ($OPT{dict}) {
	@dict = file_to_array($OPT{dict});
    }

    my @sink = ();
    if ($OPT{sink}) {
	@sink = file_to_array($OPT{sink});
    }

    while (scalar @ARGV) {
	my $data_file = shift @ARGV;
	my $data_parser = new XML::DOM::Parser;
	my $data_doc = $data_parser->parsefile($data_file);
	my $data_root = $data_doc->getDocumentElement;

	my @removed = ();
	if ($OPT{diff}) {
	    my $diff_file = shift @ARGV;
	    my $diff_parser = new XML::DOM::Parser;
	    my $diff_doc = $diff_parser->parsefile($diff_file);
	    my $diff_root = $diff_doc->getDocumentElement;
	    
	    foreach my $r ($diff_root->findnodes("//removed/\@word_id")) {
		push @removed, $r->getNodeValue;
	    }
	    
	    $diff_doc->dispose;
	}

	my @examples_pos = positives($OPT{netype}, $data_root);
	my @examples_neg = negatives($OPT{full},$OPT{negative},\@removed,\@anti,\@sink,$data_root);
	
	if ($OPT{data}) {
	    print "% $data_file\n";
	    if ($OPT{vocabulary}) {
		foreach my $ne (@examples_pos) {
		    vocabulary_data("ne", $OPT{range}, $ne);
		}
		
		foreach my $word (@examples_neg) {
		    vocabulary_data("not_ne", $OPT{range}, $word);
		}
	    }
	    else {
		foreach my $ne (@examples_pos) {
		    example_data("ne", $OPT{range}, $OPT{title}, \@dict, \@anti, \@following, \@preceding,  $ne);
		}
		
		foreach my $word (@examples_neg) {
		    example_data("not_ne", $OPT{range}, $OPT{title}, \@dict, \@anti, \@following, \@preceding, $word);
		    my $prev = $word->getPreviousSibling;
		    if ($prev && ($prev->getNodeType == TEXT_NODE) && ($prev->getNodeValue !~ /[^ \n\t]/)) {
			$prev = $prev->getPreviousSibling;
			if ($prev && ($prev->getNodeType == ELEMENT_NODE) && ($prev->getNodeName eq "word")) {
			    example_data("not_ne", $OPT{range}, $OPT{title}, \@dict, \@anti, \@following, \@preceding, $prev, $word);
			    my $even = $prev->getPreviousSibling;
			    if ($even && ($even->getNodeType == TEXT_NODE) && ($even->getNodeValue !~ /[^ \n\t]/)) {
				$even = $even->getPreviousSibling;
				if ($even && ($even->getNodeType == ELEMENT_NODE) && ($even->getNodeName eq "word")) {
				    example_data("not_ne", $OPT{range}, $OPT{title}, \@dict, \@anti, \@following, \@preceding, $even, $prev, $word);
				}
			    }
			}
		    }
		}
	    }
	}
	
	$data_doc->dispose;
	$count_examples{positive} += $#examples_pos + 1;
	$count_examples{negative} += $#examples_neg + 1;
    }

    print STDERR "Total:" . $count_examples{positive} . " positives, " . $count_examples{negative} . " negatives\n";
}
